

#include <cstdlib>
#include <cstdio>
#include <GL/glut.h>
#include <vector>
#include <fstream>
#include <cmath>
#include <iostream>
#include "NumberGenerator.h"

#define MAX_SAMPLES 256
#define PI 3.14159265

using namespace::std;

struct Point
{
	float x,y,z;

	Point() {}

	Point(float a,float b,float c)
	{
		x = a;
		y = b;
		z = c;
	}


	float Magnitude()
	{
		return sqrt(x*x + y*y + z*z);
	}

	Point Normalize()
	{
		float mag = sqrt(x*x + y*y + z*z);
		x /= mag;
		y /= mag;
		z /= mag;
		return *this;
	}

	Point operator+(Point &p)
	{
		return Point(x+p.x,y+p.y,z+p.z);		
	}

	Point operator-(Point &p)
	{
		return Point(x-p.x,y-p.y,z-p.z);		
	}

	Point operator*(float a)
	{
		return Point(x*a,y*a,z*a);
	}
};

struct Triangle
{
	Point v0,v1,v2; // vertices
	float r,g,b; // color
	Point n; // normal
	Point tangent,bitangent;
};


struct Light
{
	Point v0,v1,v2; // vertices
	float r,g,b; // color
	Point n; // normal
	Point tangent,bitangent;
	Point samples[MAX_SAMPLES];
};

Point MakeVector(Point a,Point b)
{
	Point p(0,0,0);
	p.x = b.x - a.x;
	p.y = b.y - a.y;
	p.z = b.z - a.z;
	return p;
}

Point CrossProduct(Point a, Point b)
{
	Point p;
	p.x  = a.y*b.z - a.z*b.y;
	p.y  = a.z*b.x - a.x*b.z;
	p.z  =  a.x*b.y - a.y*b.x;
	return p;
}

float DotProduct(Point a,Point b)
{	
	return a.x*b.x + a.y*b.y +a.z*b.z;	
}

int numSamples = 128;
int numTriangles,numLights;
vector<Triangle> triangles;
vector<Light> lights;
vector<Point> addedLights;
vector<Point> interPoints;


// Copyright 2001, softSurfer (www.softsurfer.com)
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.
bool RayTriangleIntersection( Point P0, Point direction, Triangle T, Point* I)
{
	Point    u, v, n;             // triangle vectors
	Point    dir, w0, w;          // ray vectors
	float     r, a, b;             // params to calc ray-plane intersect

	// get triangle edge vectors and plane normal
	n = T.n;             // cross product

	dir = direction;             // ray direction vector
	u = MakeVector(T.v0,T.v1);
	v = MakeVector(T.v0,T.v2);
	w0 = MakeVector(T.v0,P0);
	a = -DotProduct(n,w0);
	b = DotProduct(n,dir);
	if (fabs(b) < 1e-8) {     // ray is parallel to triangle plane
		if (a == 0)                // ray lies in triangle plane
			return false;
		else return false;             // ray disjoint from plane
	}

	// get intersect point of ray with triangle plane
	r = a / b;
	if (r < 0.0)                   // ray goes away from triangle
		return 0;                  // => no intersect
	// for a segment, also test if (r > 1.0) => no intersect

	dir.x*=r;
	dir.y*=r;
	dir.z*=r;

	*I = P0 + dir;           // intersect point of ray and plane

	// is I inside T?
	float    uu, uv, vv, wu, wv, D;
	uu = DotProduct(u,u);
	uv = DotProduct(u,v);
	vv = DotProduct(v,v);
	w = *I - T.v0;
	wu = DotProduct(w,u);
	wv = DotProduct(w,v);
	D = uv * uv - uu * vv;

	// get and test parametric coords
	float s, t;
	s = (uv * wv - vv * wu) / D;
	if (s < 0.0 || s > 1.0)        // I is outside T
		return false;
	t = (uv * wu - uu * wv) / D;
	if (t < 0.0 || (s + t) > 1.0)  // I is outside T
		return false;

	return true;                      // I is in T
}

void
render(void)
{

	glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT); 

	glPushMatrix();	
	/*
	** Since we want to turn texturing on for floor only, we have to
	** make floor a separate glBegin()/glEnd() sequence. You can't
	** turn texturing on and off between begin and end calls
	*/

	for (int i=0;i<numTriangles;i++)
	{
		glBegin(GL_TRIANGLES);

		glNormal3f(triangles[i].n.x,triangles[i].n.y,triangles[i].n.z);
		
		glColor3f(triangles[i].r,triangles[i].g,triangles[i].b);
		glVertex3f(triangles[i].v0.x,triangles[i].v0.y,triangles[i].v0.z);				
		glVertex3f(triangles[i].v1.x,triangles[i].v1.y,triangles[i].v1.z);			
		glVertex3f(triangles[i].v2.x,triangles[i].v2.y,triangles[i].v2.z);

		glEnd();
	} 
	glPopMatrix();

	glPointSize(5.0);
	for (int i=0;i<addedLights.size()-1;i++)
	{
//	glBegin(GL_POINTS);
	glBegin(GL_LINES);
	glColor3f(0.0,0.0,1.0);	
		glVertex3f(addedLights[i].x,addedLights[i].y,addedLights[i].z);
		glVertex3f(interPoints[i].x,interPoints[i].y,interPoints[i].z);
	glEnd();
	}
	


	/*
	glDisable(GL_TEXTURE_2D);	

	glBegin(GL_QUADS);

	glNormal3f(1.f, 0.f, 0.f);
	glVertex3f(-100.f, -100.f, -320.f);
	glVertex3f(-100.f, -100.f, -520.f);
	glVertex3f(-100.f,  100.f, -520.f);
	glVertex3f(-100.f,  100.f, -320.f);


	glNormal3f(-1.f, 0.f, 0.f);
	glVertex3f( 100.f, -100.f, -320.f);
	glVertex3f( 100.f,  100.f, -320.f);
	glVertex3f( 100.f,  100.f, -520.f);
	glVertex3f( 100.f, -100.f, -520.f);

	glNormal3f(0.f, -1.f, 0.f);
	glVertex3f(-100.f,  100.f, -320.f);
	glVertex3f(-100.f,  100.f, -520.f);
	glVertex3f( 100.f,  100.f, -520.f);
	glVertex3f( 100.f,  100.f, -320.f);

	glNormal3f(0.f, 0.f, 1.f);
	glVertex3f(-100.f, -100.f, -520.f);
	glVertex3f( 100.f, -100.f, -520.f);
	glVertex3f( 100.f,  100.f, -520.f);
	glVertex3f(-100.f,  100.f, -520.f);
	glEnd();
	*/

	if(glGetError()) /* to catch programming errors; should never happen */
		printf("Oops! I screwed up my OpenGL calls somewhere\n");

	glFlush(); /* high end machines may need this */
}

enum {NONE, AA};

int rendermode = NONE;

void
menu(int selection)
{
	rendermode = selection;
	glutPostRedisplay();
}

/* Called when window needs to be redrawn */
void redraw()
{
	int numMissed = 0;
	addedLights.clear();
	interPoints.clear();
	float rho = 0.5774;
	//float p[] = {5000,300.8,600.2};
	float p[] = { -500,-250,-100};
	float inten[] = { 1.0, 1.0, 1.0, 1.0};
	//float rho = 0.85;

	switch(rendermode) {
case NONE:

	glClear(GL_ACCUM_BUFFER_BIT);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60,1,1,5000);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt( 278.0,
		273.0,
		-800.0,
		278.0,
		273.0,
		800.0,
		0.0,
		1.0,
		0.0);	
	

	// Keller[99] algorithm

	for(int k = 0; k < numLights; k++)
	{
		int N = numSamples/2;

		//Algorithm straight out of the paper
		float w, start;
		int end, reflections = 0;		
		Point y, omega;

		start = end = N;

		while(end > 0)
		{
			start *= rho;

			for(int i = (int)start; i < end; i++)
			{

				Point y = lights[k].samples[i];
				Point normal = lights[k].n;
				Point tangent = lights[k].tangent;
				Point bitangent = lights[k].bitangent;
				
				float lightColor[4];

				lightColor[0] = lights[k].r;
				lightColor[1] = lights[k].g;
				lightColor[2] = lights[k].b;
				lightColor[3] = 0.5;

				w = N;

				//trace reflections
				for(int j = 0; j <= reflections; j++)
				{						
					float dist  = -1;
					float pos[4];
					float amb[] = {0.0,0.0,0.0,0.0};

					addedLights.push_back(y);

					pos[0] = y.x;
					pos[1]= y.y;
					pos[2] = y.z;
					pos[3] = 1.0;
					
					//cout<<"Reflection:"<<j<<" Added light at:"<<y.x<<" "<<y.y<<" "<<y.z<<endl;
					//Place node in correct position					
					

					for (int count = 0;count<4;count++)
						lightColor[count] *= (N*1.0)/(floorf(w));

					glLightfv(GL_LIGHT0, GL_DIFFUSE, lightColor);	
					glLightfv(GL_LIGHT0, GL_POSITION, pos);  																	
										
					// Render with new light source

					render();
					glAccum(GL_ACCUM, 1.0f/(1.0*numSamples)); 
			

					// Get directions
					std::vector<Vector2> results = HaltonGenerator::getInstance().getPairs(numSamples, 2*j+2, 2*j+3); 

					//std::vector<Vector2> results = UniformGenerator::getInstance().getPairs();

					float zenith = asin(sqrt(results[i].x));
					float azimuth = results[i].y * 2 * PI;

					//I use the tangent as my x axis
					Point projectedDirection(cos(azimuth), sin(azimuth), 0.0);

					//I use the normal as my z axis
					float cosZenith = cos(zenith);
					//Point direction(cosZenith * projectedDirection.x, cosZenith * projectedDirection.y, sin(zenith));

					Point direction(projectedDirection.x,projectedDirection.y,projectedDirection.Magnitude()* tan(zenith));

					//just in case
					direction = tangent*direction.x + bitangent*direction.y +  normal*direction.z;
					direction.Normalize();


					bool intersection = false;					
					Point intersectionPoint,tempIntersectionPoint;
					int intersectionTriangleIndex;


					for (int count =0; count<numTriangles;count++)
					{
						if (RayTriangleIntersection(y,direction,triangles[count],&tempIntersectionPoint))
						{
							if (MakeVector(y,tempIntersectionPoint).Magnitude()<dist || dist == -1)
							{
								dist = MakeVector(y,tempIntersectionPoint).Magnitude();
								if (dist == 0) continue;
								intersectionPoint = tempIntersectionPoint;
								intersectionTriangleIndex = count;
								intersection = true;
							}
						}
					}

					// Cast a ray, get intersection point and triangle
					if (intersection)
					{											
						interPoints.push_back(intersectionPoint);
						y = intersectionPoint;
					//	cout<<"Intersection at:"<<y.x<<" "<<y.y<<" "<<y.z<<endl;
						normal = triangles[intersectionTriangleIndex].n;
						tangent = triangles[intersectionTriangleIndex].tangent;
						bitangent = triangles[intersectionTriangleIndex].bitangent;

						lightColor[0] *= triangles[intersectionTriangleIndex].r;
						lightColor[1] *= triangles[intersectionTriangleIndex].g;
						lightColor[2] *= triangles[intersectionTriangleIndex].b;						

						w *= rho;
					}			
					else 
					{
						numMissed++;
						//cout<<"No intersection"<<endl;
						interPoints.push_back(y);
					}
				}

			}
			reflections++;		
			end = (int)start;
		}
	}

	glAccum(GL_RETURN, 1.f);	
//	cout<<"Number of missed rays:"<<numMissed<<endl;
	break;

case AA:




	break;
	}

	glutSwapBuffers();
}

void key(unsigned char key, int x, int y)
{
	if(key == '\033')
		exit(0);
}


const int TEXDIM = 256;

/* Parse arguments, and set up interface between OpenGL and window system */
void main(int argc, char *argv[])
{
	glutInit(&argc, argv);
	glutInitWindowSize(800, 600);
	glutInitDisplayMode(GLUT_RGBA|GLUT_DEPTH|GLUT_ACCUM|GLUT_SINGLE);
	(void)glutCreateWindow("GL Instant Radiosity");
	glutDisplayFunc(redraw);
	glutKeyboardFunc(key);

	glutCreateMenu(menu);
	glutAddMenuEntry("Aliased View", NONE);
	glutAddMenuEntry("AntiAliased", AA);
	glutAttachMenu(GLUT_RIGHT_BUTTON);



	/* turn on features */
	//GLfloat global_ambient[] = { 0.01f, 0.01f, 0.01f, 1.0f };
	//glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);

/*	GLfloat pos[] = { 530.0, 0.0, 500.0, 0.0};
	GLfloat col[] = { 1.0, 1.0, 1.0, 1.0};
	GLfloat amb[] = {0.0,0.0,0.0,0.0};
	glLightfv(GL_LIGHT0,GL_POSITION,pos);
	glLightfv(GL_LIGHT0,GL_DIFFUSE,col);
	glLightfv(GL_LIGHT0,GL_SPECULAR,col);
	glLightfv(GL_LIGHT0,GL_AMBIENT,amb); */
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);

	glColorMaterial ( GL_FRONT_AND_BACK, GL_DIFFUSE) ;
	glEnable ( GL_COLOR_MATERIAL ) ;
	glEnable(GL_LIGHT0);					

	glClearColor(0.0,0.0,0.0,0.0);

	/* remove back faces to speed things up */
	glCullFace(GL_BACK);

	// Read in triangle data

	fstream myFile;
	myFile.open ("f:\\mesh.txt",ios::in);

	numTriangles = 0; 
	numLights = 0;
	int numQuads;
	while (!myFile.eof())
	{
		float r,g,b;
		myFile >> numQuads;
		myFile >> r >> g >> b;
		for (int i=0;i<numQuads;i++)
		{
			Triangle t0,t1;
			Point normal;
			Point vec1,vec2;
			Point quad[4];

			for (int j=0;j<4;j++) 
				myFile >> quad[j].x >> quad[j].y >> quad[j].z;

			t0.v0 = quad[0];
			t0.v1 = quad[1];
			t0.v2 = quad[2];

			vec1 = MakeVector(t0.v0,t0.v1);
			vec2 = MakeVector(t0.v0,t0.v2);
			normal = CrossProduct(vec1,vec2);
			normal.Normalize();
			
			t0.n = normal;
			t0.tangent = vec1.Normalize();
			t0.bitangent = CrossProduct(t0.n,t0.tangent).Normalize();
			t0.r = r;
			t0.g = g;
			t0.b = b;
			triangles.push_back(t0);

			// 2nd triangle
			t1.v0 = quad[0];
			t1.v1 = quad[2];
			t1.v2 = quad[3];

			vec1 = MakeVector(t1.v0,t1.v1);
			vec2 = MakeVector(t1.v0,t1.v2);
			normal = CrossProduct(vec1,vec2);
			normal.Normalize();

			t1.n = normal;
			t1.tangent = vec1.Normalize();
			t1.bitangent = CrossProduct(t1.n,t1.tangent).Normalize();
			t1.r = r;
			t1.g = g;
			t1.b = b;

			triangles.push_back(t1);

			numTriangles+=2;
		}
	}
	myFile.close();

	// Read lights

	myFile.clear();
	myFile.open ("f:\\light.txt",ios::in);
	

	while (!myFile.eof())
	{
		float r,g,b;
		myFile >> numQuads;
		myFile >> r >> g >> b;
		for (int i=0;i<numQuads;i++)
		{
			Light t0,t1;
			Point normal;
			Point vec1,vec2;
			Point quad[4];

			for (int j=0;j<4;j++) 
				myFile >> quad[j].x >> quad[j].y >> quad[j].z;

			t0.v0 = quad[0];
			t0.v1 = quad[1];
			t0.v2 = quad[2];

			vec1 = MakeVector(t0.v0,t0.v1);
			vec2 = MakeVector(t0.v0,t0.v2);
			normal = CrossProduct(vec1,vec2);
			normal.Normalize();

			t0.n = normal;
			t0.tangent = vec1.Normalize();
			t0.bitangent = CrossProduct(t0.n,t0.tangent).Normalize();

			t0.r = r;
			t0.g = g;
			t0.b = b;

			std::vector<Vector2> results;
			results = HaltonGenerator::getInstance().getPairs(numSamples, 1, 2);
//			results = UniformGenerator::getInstance().getPairs();

			for (int sampleCount = 0; sampleCount < numSamples; sampleCount++)
			{
				float u = results[sampleCount].x;
				float v = results[sampleCount].y;

				float s = 1 - sqrt(1-u);
				float t = (1-s)*v;


				t0.samples[sampleCount] = t0.v0 +
					(t0.v1 - t0.v0)*s +
					(t0.v2 - t0.v0)*t;
			}
			lights.push_back(t0);

			// 2nd triangle
			t1.v0 = quad[0];
			t1.v1 = quad[2];
			t1.v2 = quad[3];

			vec1 = MakeVector(t1.v0,t1.v1);
			vec2 = MakeVector(t1.v0,t1.v2);
			normal = CrossProduct(vec1,vec2);
			normal.Normalize();

			t1.n = normal;
			t1.tangent = vec1.Normalize();
			t1.bitangent = CrossProduct(t1.n,t1.tangent).Normalize();

			t1.r = r;
			t1.g = g;
			t1.b = b;

			results = HaltonGenerator::getInstance().getPairs(numSamples, 1, 2);
//			results = UniformGenerator::getInstance().getPairs();

			for (int sampleCount = 0; sampleCount < numSamples; sampleCount++)
			{
				float u = results[sampleCount].x;
				float v = results[sampleCount].y;

				float s = 1 - sqrt(1-u);
				float t = (1-s)*v;


				t1.samples[sampleCount] = t1.v0 +
					 (t1.v1 - t1.v0)*s +
					(t1.v2 - t1.v0)*t;
			}

			lights.push_back(t1);

			numLights+=2;
		}
		break;
	}
	myFile.close();

	glReadBuffer(GL_BACK); /* input to accum buffer */

	glutMainLoop();
}
